from . import entity, generator, llm, prompt


class GatoService:
    _scenario_prompt_factory = prompt.ScenarioPromptFactory()
    _action_prompt_factory = prompt.ActionPromptFactory()

    def __init__(self, model: llm.LLM):
        self._model = model

    def create_scenario(self) -> entity.Scenario:
        scenario_prompt = self._scenario_prompt_factory.new()
        scenario_generator = generator.ScenarioGenerator(self._model, scenario_prompt)
        return scenario_generator.generate()

    def create_action(self, scenario: entity.Scenario) -> entity.Action:
        action_prompt = self._action_prompt_factory.new(scenario)
        action_generator = generator.ActionGenerator(self._model, action_prompt)
        return action_generator.generate()



import pathlib
import random

from . import entity


def read_file(path):
    with open(path, encoding="utf-8", errors="ignore") as file:
        return file.read()


def read_list(path):
    return read_file(path).strip().split('\n')


class ScenarioParametersFactory:
    _base_dir = pathlib.Path(__file__).parent
    _default_path = pathlib.Path(".config/scenario-params")

    def __init__(self, path: pathlib.Path = _default_path):
        self._path = path

    def _get_random_parameter(self, name: str):
        path = self._base_dir / self._path / f"{name}.txt"
        choices = read_list(path)
        return random.choice(choices)

    def random(self) -> entity.ScenarioParameters:
        return entity.ScenarioParameters(
            random_word=self._get_random_parameter("entropies"),
            scope=self._get_random_parameter("scopes"),
            severity=self._get_random_parameter("severities"),
            region=self._get_random_parameter("regions"),
            category=self._get_random_parameter("categories"),
            domain=self._get_random_parameter("domains"),
        )


class ScenarioPromptFactory:
    _base_dir = pathlib.Path(__file__).parent
    _default_path = pathlib.Path(".config/scenario-system-messages.txt")
    _parameters_factory = ScenarioParametersFactory()

    def __init__(self,
                 parameters: entity.ScenarioParameters = _parameters_factory.random(),
                 path: pathlib.Path = _base_dir / _default_path):
        self._parameters = parameters
        self._system_message = random.choice(read_list(path))

    def new(self) -> entity.ScenarioPrompt:
        return entity.ScenarioPrompt(
            system_message=self._system_message,
            parameters=self._parameters,
        )


class ActionPromptFactory:
    _base_dir = pathlib.Path(__file__).parent
    _default_path = pathlib.Path(".config/scenario-params")

    def __init__(self, path: pathlib.Path = _default_path):
        self._path = path

    def new(self, scenario: entity.Scenario) -> entity.ActionPrompt:
        system_message = read_file(self._path)
        return entity.ActionPrompt(
            system_message=system_message,
            scenario=scenario,
        )



import time

import openai as openai


class ChatPrompt:
    def __init__(self):
        self._messages: list[dict[str, str]] = []

    def add_message(self, role: str, content: str):
        self._messages.append({"role": role, "content": content})

    def add_system_message(self, content: str):
        self.add_message("system", content)

    def add_user_message(self, content: str):
        self.add_message("user", content)

    def messages(self):
        return self._messages


class LLM:
    def __init__(self, api_key, model_name: str = "gpt-3.5-turbo"):
        openai.api_key = api_key
        self._model_name = model_name

    def _query(self, prompt: ChatPrompt, **kwargs) -> str:
        response = openai.ChatCompletion.create(
            model=self._model_name,
            messages=prompt.messages(),
            **kwargs,
        )
        return response.choices[0]["message"]["content"]

    def submit(self, prompt: ChatPrompt, temperature: float = 0.7,
               max_retries: int = 3, **kwargs) -> str:
        for retry in range(max_retries):
            try:
                return self._query(prompt, temperature=temperature, **kwargs)
            except Exception as err:
                wait = 2 ** retry * 5
                print(f'Error communicating with OpenAI: "{err}"'
                      f' - Retrying in {wait} seconds...')
                time.sleep(wait)
        print(f"Failed after {max_retries} queries - aborting.")



from . import entity, llm


class ScenarioGenerator:

    def __init__(self, model: llm.LLM, prompt: entity.ScenarioPrompt):
        self._model = model
        self._prompt = prompt

    def generate(self) -> entity.Scenario:
        system_message = self._prompt.system_message
        parameters = self._prompt.parameters.dict().items()
        user_message = "\n".join([f"{name}: {param}" for name, param in parameters])

        prompt = llm.ChatPrompt()
        prompt.add_system_message(system_message)
        prompt.add_user_message(user_message)
        description = self._model.submit(prompt)
        return entity.Scenario(description=description)


class ActionGenerator:

    def __init__(self, model: llm.LLM, prompt: entity.ActionPrompt):
        self._model = model
        self._prompt = prompt

    def generate(self) -> entity.Action:
        system_message = self._prompt.system_message
        scenario_description = self._prompt.scenario.description

        prompt = llm.ChatPrompt()
        prompt.add_system_message(system_message)
        prompt.add_user_message(scenario_description)
        description = self._model.submit(prompt)
        return entity.Action(description=description)



import uuid

from pydantic import BaseModel, Field


class ScenarioParameters(BaseModel):
    random_word: str
    scope: str
    severity: str
    region: str
    category: str
    domain: str


class ScenarioPrompt(BaseModel):
    system_message: str
    parameters: ScenarioParameters


class Scenario(BaseModel):
    description: str
    id: uuid.UUID = Field(default_factory=uuid.uuid4)


class ActionPrompt(BaseModel):
    system_message: str
    scenario: Scenario


class Action(BaseModel):
    description: str
    id: uuid.UUID = Field(default_factory=uuid.uuid4)



from setuptools import setup, find_packages


def get_long_description():
    with open('README.md') as file:
        return file.read()


setup(
    name="gato-toolkit",
    version="0.1.0",
    description="A toolkit for furthering research on AI alignment.",
    url="https://github.com/FyZyX/gato-toolkit",
    author="Lucas Lofaro <lucasmlofaro@gmail.com>",
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    include_package_data=True,
    license='MIT',
    long_description=get_long_description(),
    long_description_content_type='text/markdown',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.10',
    ],
    install_requires=[
        "openai>=0.27.6",
        "pydantic>=1.10.7",
    ],
)
